## Dynamic C-code Library.
useDynLib(randomForestRHF, .registration = TRUE)

importFrom("parallel", mclapply)
importFrom("randomForestSRC", "rfsrc")

importFrom("grDevices", "adjustcolor")

importFrom("graphics", "legend", "lines", "par", "rug", "mtext", "points")
importFrom("graphics", "matplot", "boxplot", "abline", "polygon")


importFrom("stats", "runif", "as.formula", "formula", "na.omit", "model.matrix", "sd", "var", "supsmu")
importFrom("stats", "quantile", "rbinom", "setNames", "lowess", "uniroot")
importFrom("stats", "complete.cases")
importFrom("stats", "plogis")

importFrom("survival", "Surv")

importFrom("utils", "tail", "combn", "capture.output", "head", "flush.console")

importFrom("varPro", "varpro", "get.orgvimp", "get.vimp", "cv.varpro",
           "varpro.strength", "get.varpro.strengthArray",
           "local.importance", "importance.varpro.workhorse")

export(as.data.frame.importance.rhf,
       auct.rhf,
       auct,
       convert.counting,
       convert.standard.counting,
       dotmatrix.importance.rhf,
       dotmatrix.importance,
       hazard.simulation,
       importance.rhf,
       plot.auct.rhf,
       plot.importance.rhf,
       plot.rhf,
       plot.tune.treesize.rhf,
       predict.rhf,
       print.auct.rhf,
       print.importance.rhf,
       print.rhf,       
       rhf,
       rhf.news,
       smoothed.hazard.rhf,
       smoothed.hazard,
       tune.treesize.rhf,
       tune.rhf,
       tune.iAUC.rhf,
       tune.iAUC,
       varpro.cache.rhf,
       varpro.cache,
       xvar.wt.rhf)

S3method(as.data.frame, importance.rhf)

S3method(plot, auct.rhf)
S3method(plot, importance.rhf)
S3method(plot, rhf)
S3method(plot, tune.treesize.rhf)


S3method(predict, rhf)

S3method(print, auct.rhf)
S3method(print, importance.rhf)
S3method(print, rhf)


